# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/batch_data_source.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.args import data_source_pb2 as tecton__proto_dot_args_dot_data__source__pb2
from tecton_proto.args import data_source_config_pb2 as tecton__proto_dot_args_dot_data__source__config__pb2
from tecton_proto.args import user_defined_function_pb2 as tecton__proto_dot_args_dot_user__defined__function__pb2
from tecton_proto.common import spark_schema_pb2 as tecton__proto_dot_common_dot_spark__schema__pb2
from tecton_proto.data import hive_metastore_pb2 as tecton__proto_dot_data_dot_hive__metastore__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/data/batch_data_source.proto',
  package='tecton_proto.data',
  syntax='proto2',
  serialized_options=b'\n\017com.tecton.dataP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n)tecton_proto/data/batch_data_source.proto\x12\x11tecton_proto.data\x1a#tecton_proto/args/data_source.proto\x1a*tecton_proto/args/data_source_config.proto\x1a-tecton_proto/args/user_defined_function.proto\x1a&tecton_proto/common/spark_schema.proto\x1a&tecton_proto/data/hive_metastore.proto\"T\n\x19TimestampColumnProperties\x12\x1f\n\x0b\x63olumn_name\x18\x01 \x01(\tR\ncolumnName\x12\x16\n\x06\x66ormat\x18\x02 \x01(\tR\x06\x66ormat\"\xa9\x06\n\x0f\x42\x61tchDataSource\x12G\n\nhive_table\x18\x01 \x01(\x0b\x32&.tecton_proto.data.HiveTableDataSourceH\x00R\thiveTable\x12\x37\n\x04\x66ile\x18\x08 \x01(\x0b\x32!.tecton_proto.data.FileDataSourceH\x00R\x04\x66ile\x12H\n\x0bredshift_db\x18\x0b \x01(\x0b\x32%.tecton_proto.data.RedshiftDataSourceH\x00R\nredshiftDb\x12\x46\n\tsnowflake\x18\x0c \x01(\x0b\x32&.tecton_proto.data.SnowflakeDataSourceH\x00R\tsnowflake\x12\x43\n\x0cspark_schema\x18\t \x01(\x0b\x32 .tecton_proto.common.SparkSchemaR\x0bsparkSchema\x12l\n\x1btimestamp_column_properties\x18\x04 \x01(\x0b\x32,.tecton_proto.data.TimestampColumnPropertiesR\x19timestampColumnProperties\x12\x41\n\x0c\x62\x61tch_config\x18\x05 \x01(\x0b\x32\x1e.tecton_proto.args.BatchConfigR\x0b\x62\x61tchConfig\x12\x32\n\x15\x64\x61te_partition_column\x18\x06 \x01(\tR\x13\x64\x61tePartitionColumn\x12h\n\x1a\x64\x61tetime_partition_columns\x18\x07 \x03(\x0b\x32*.tecton_proto.data.DatetimePartitionColumnR\x18\x64\x61tetimePartitionColumns\x12X\n\x14raw_batch_translator\x18\n \x01(\x0b\x32&.tecton_proto.args.UserDefinedFunctionR\x12rawBatchTranslatorB\x0e\n\x0c\x62\x61tch_sourceJ\x04\x08\x02\x10\x03\"\x88\x02\n\x0e\x46ileDataSource\x12\x10\n\x03uri\x18\x01 \x02(\tR\x03uri\x12?\n\x06\x66ormat\x18\x02 \x02(\x0e\x32\'.tecton_proto.data.FileDataSourceFormatR\x06\x66ormat\x12\x33\n\x16\x63onvert_to_glue_format\x18\x04 \x01(\x08R\x13\x63onvertToGlueFormat\x12\x1d\n\nschema_uri\x18\x05 \x01(\tR\tschemaUri\x12I\n\x0fschema_override\x18\x06 \x01(\x0b\x32 .tecton_proto.common.SparkSchemaR\x0eschemaOverrideJ\x04\x08\x03\x10\x04\"\x88\x01\n\x17\x44\x61tetimePartitionColumn\x12\x1f\n\x0b\x63olumn_name\x18\x01 \x02(\tR\ncolumnName\x12#\n\rformat_string\x18\x02 \x02(\tR\x0c\x66ormatString\x12\'\n\x0fminimum_seconds\x18\x03 \x02(\x03R\x0eminimumSeconds\"\xbe\x01\n\x12RedshiftDataSource\x12\x1a\n\x08\x65ndpoint\x18\x01 \x01(\tR\x08\x65ndpoint\x12\x1d\n\ncluster_id\x18\x02 \x01(\tR\tclusterId\x12\x1a\n\x08\x64\x61tabase\x18\x03 \x01(\tR\x08\x64\x61tabase\x12\x16\n\x05table\x18\x04 \x01(\tH\x00R\x05table\x12\x16\n\x05query\x18\x06 \x01(\tH\x00R\x05query\x12\x17\n\x07temp_s3\x18\x05 \x01(\tR\x06tempS3B\x08\n\x06source\"g\n\x13SnowflakeDataSource\x12P\n\rsnowflakeArgs\x18\x01 \x01(\x0b\x32*.tecton_proto.args.SnowflakeDataSourceArgsR\rsnowflakeArgs*~\n\x14\x46ileDataSourceFormat\x12 \n\x1c\x46ILE_DATA_SOURCE_FORMAT_JSON\x10\x00\x12#\n\x1f\x46ILE_DATA_SOURCE_FORMAT_PARQUET\x10\x01\x12\x1f\n\x1b\x46ILE_DATA_SOURCE_FORMAT_CSV\x10\x02\x42\x13\n\x0f\x63om.tecton.dataP\x01'
  ,
  dependencies=[tecton__proto_dot_args_dot_data__source__pb2.DESCRIPTOR,tecton__proto_dot_args_dot_data__source__config__pb2.DESCRIPTOR,tecton__proto_dot_args_dot_user__defined__function__pb2.DESCRIPTOR,tecton__proto_dot_common_dot_spark__schema__pb2.DESCRIPTOR,tecton__proto_dot_data_dot_hive__metastore__pb2.DESCRIPTOR,])

_FILEDATASOURCEFORMAT = _descriptor.EnumDescriptor(
  name='FileDataSourceFormat',
  full_name='tecton_proto.data.FileDataSourceFormat',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='FILE_DATA_SOURCE_FORMAT_JSON', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FILE_DATA_SOURCE_FORMAT_PARQUET', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FILE_DATA_SOURCE_FORMAT_CSV', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1874,
  serialized_end=2000,
)
_sym_db.RegisterEnumDescriptor(_FILEDATASOURCEFORMAT)

FileDataSourceFormat = enum_type_wrapper.EnumTypeWrapper(_FILEDATASOURCEFORMAT)
FILE_DATA_SOURCE_FORMAT_JSON = 0
FILE_DATA_SOURCE_FORMAT_PARQUET = 1
FILE_DATA_SOURCE_FORMAT_CSV = 2



_TIMESTAMPCOLUMNPROPERTIES = _descriptor.Descriptor(
  name='TimestampColumnProperties',
  full_name='tecton_proto.data.TimestampColumnProperties',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='column_name', full_name='tecton_proto.data.TimestampColumnProperties.column_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='columnName', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='format', full_name='tecton_proto.data.TimestampColumnProperties.format', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='format', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=272,
  serialized_end=356,
)


_BATCHDATASOURCE = _descriptor.Descriptor(
  name='BatchDataSource',
  full_name='tecton_proto.data.BatchDataSource',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='hive_table', full_name='tecton_proto.data.BatchDataSource.hive_table', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='hiveTable', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='file', full_name='tecton_proto.data.BatchDataSource.file', index=1,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='file', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='redshift_db', full_name='tecton_proto.data.BatchDataSource.redshift_db', index=2,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='redshiftDb', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='snowflake', full_name='tecton_proto.data.BatchDataSource.snowflake', index=3,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='snowflake', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='spark_schema', full_name='tecton_proto.data.BatchDataSource.spark_schema', index=4,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='sparkSchema', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp_column_properties', full_name='tecton_proto.data.BatchDataSource.timestamp_column_properties', index=5,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='timestampColumnProperties', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='batch_config', full_name='tecton_proto.data.BatchDataSource.batch_config', index=6,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='batchConfig', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='date_partition_column', full_name='tecton_proto.data.BatchDataSource.date_partition_column', index=7,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='datePartitionColumn', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='datetime_partition_columns', full_name='tecton_proto.data.BatchDataSource.datetime_partition_columns', index=8,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='datetimePartitionColumns', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='raw_batch_translator', full_name='tecton_proto.data.BatchDataSource.raw_batch_translator', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='rawBatchTranslator', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='batch_source', full_name='tecton_proto.data.BatchDataSource.batch_source',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=359,
  serialized_end=1168,
)


_FILEDATASOURCE = _descriptor.Descriptor(
  name='FileDataSource',
  full_name='tecton_proto.data.FileDataSource',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='uri', full_name='tecton_proto.data.FileDataSource.uri', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='uri', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='format', full_name='tecton_proto.data.FileDataSource.format', index=1,
      number=2, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='format', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='convert_to_glue_format', full_name='tecton_proto.data.FileDataSource.convert_to_glue_format', index=2,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='convertToGlueFormat', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema_uri', full_name='tecton_proto.data.FileDataSource.schema_uri', index=3,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='schemaUri', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema_override', full_name='tecton_proto.data.FileDataSource.schema_override', index=4,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='schemaOverride', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1171,
  serialized_end=1435,
)


_DATETIMEPARTITIONCOLUMN = _descriptor.Descriptor(
  name='DatetimePartitionColumn',
  full_name='tecton_proto.data.DatetimePartitionColumn',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='column_name', full_name='tecton_proto.data.DatetimePartitionColumn.column_name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='columnName', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='format_string', full_name='tecton_proto.data.DatetimePartitionColumn.format_string', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='formatString', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='minimum_seconds', full_name='tecton_proto.data.DatetimePartitionColumn.minimum_seconds', index=2,
      number=3, type=3, cpp_type=2, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='minimumSeconds', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1438,
  serialized_end=1574,
)


_REDSHIFTDATASOURCE = _descriptor.Descriptor(
  name='RedshiftDataSource',
  full_name='tecton_proto.data.RedshiftDataSource',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='endpoint', full_name='tecton_proto.data.RedshiftDataSource.endpoint', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='endpoint', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cluster_id', full_name='tecton_proto.data.RedshiftDataSource.cluster_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='clusterId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='database', full_name='tecton_proto.data.RedshiftDataSource.database', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='database', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='table', full_name='tecton_proto.data.RedshiftDataSource.table', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='table', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='query', full_name='tecton_proto.data.RedshiftDataSource.query', index=4,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='query', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='temp_s3', full_name='tecton_proto.data.RedshiftDataSource.temp_s3', index=5,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='tempS3', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='source', full_name='tecton_proto.data.RedshiftDataSource.source',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=1577,
  serialized_end=1767,
)


_SNOWFLAKEDATASOURCE = _descriptor.Descriptor(
  name='SnowflakeDataSource',
  full_name='tecton_proto.data.SnowflakeDataSource',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='snowflakeArgs', full_name='tecton_proto.data.SnowflakeDataSource.snowflakeArgs', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='snowflakeArgs', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1769,
  serialized_end=1872,
)

_BATCHDATASOURCE.fields_by_name['hive_table'].message_type = tecton__proto_dot_data_dot_hive__metastore__pb2._HIVETABLEDATASOURCE
_BATCHDATASOURCE.fields_by_name['file'].message_type = _FILEDATASOURCE
_BATCHDATASOURCE.fields_by_name['redshift_db'].message_type = _REDSHIFTDATASOURCE
_BATCHDATASOURCE.fields_by_name['snowflake'].message_type = _SNOWFLAKEDATASOURCE
_BATCHDATASOURCE.fields_by_name['spark_schema'].message_type = tecton__proto_dot_common_dot_spark__schema__pb2._SPARKSCHEMA
_BATCHDATASOURCE.fields_by_name['timestamp_column_properties'].message_type = _TIMESTAMPCOLUMNPROPERTIES
_BATCHDATASOURCE.fields_by_name['batch_config'].message_type = tecton__proto_dot_args_dot_data__source__config__pb2._BATCHCONFIG
_BATCHDATASOURCE.fields_by_name['datetime_partition_columns'].message_type = _DATETIMEPARTITIONCOLUMN
_BATCHDATASOURCE.fields_by_name['raw_batch_translator'].message_type = tecton__proto_dot_args_dot_user__defined__function__pb2._USERDEFINEDFUNCTION
_BATCHDATASOURCE.oneofs_by_name['batch_source'].fields.append(
  _BATCHDATASOURCE.fields_by_name['hive_table'])
_BATCHDATASOURCE.fields_by_name['hive_table'].containing_oneof = _BATCHDATASOURCE.oneofs_by_name['batch_source']
_BATCHDATASOURCE.oneofs_by_name['batch_source'].fields.append(
  _BATCHDATASOURCE.fields_by_name['file'])
_BATCHDATASOURCE.fields_by_name['file'].containing_oneof = _BATCHDATASOURCE.oneofs_by_name['batch_source']
_BATCHDATASOURCE.oneofs_by_name['batch_source'].fields.append(
  _BATCHDATASOURCE.fields_by_name['redshift_db'])
_BATCHDATASOURCE.fields_by_name['redshift_db'].containing_oneof = _BATCHDATASOURCE.oneofs_by_name['batch_source']
_BATCHDATASOURCE.oneofs_by_name['batch_source'].fields.append(
  _BATCHDATASOURCE.fields_by_name['snowflake'])
_BATCHDATASOURCE.fields_by_name['snowflake'].containing_oneof = _BATCHDATASOURCE.oneofs_by_name['batch_source']
_FILEDATASOURCE.fields_by_name['format'].enum_type = _FILEDATASOURCEFORMAT
_FILEDATASOURCE.fields_by_name['schema_override'].message_type = tecton__proto_dot_common_dot_spark__schema__pb2._SPARKSCHEMA
_REDSHIFTDATASOURCE.oneofs_by_name['source'].fields.append(
  _REDSHIFTDATASOURCE.fields_by_name['table'])
_REDSHIFTDATASOURCE.fields_by_name['table'].containing_oneof = _REDSHIFTDATASOURCE.oneofs_by_name['source']
_REDSHIFTDATASOURCE.oneofs_by_name['source'].fields.append(
  _REDSHIFTDATASOURCE.fields_by_name['query'])
_REDSHIFTDATASOURCE.fields_by_name['query'].containing_oneof = _REDSHIFTDATASOURCE.oneofs_by_name['source']
_SNOWFLAKEDATASOURCE.fields_by_name['snowflakeArgs'].message_type = tecton__proto_dot_args_dot_data__source__pb2._SNOWFLAKEDATASOURCEARGS
DESCRIPTOR.message_types_by_name['TimestampColumnProperties'] = _TIMESTAMPCOLUMNPROPERTIES
DESCRIPTOR.message_types_by_name['BatchDataSource'] = _BATCHDATASOURCE
DESCRIPTOR.message_types_by_name['FileDataSource'] = _FILEDATASOURCE
DESCRIPTOR.message_types_by_name['DatetimePartitionColumn'] = _DATETIMEPARTITIONCOLUMN
DESCRIPTOR.message_types_by_name['RedshiftDataSource'] = _REDSHIFTDATASOURCE
DESCRIPTOR.message_types_by_name['SnowflakeDataSource'] = _SNOWFLAKEDATASOURCE
DESCRIPTOR.enum_types_by_name['FileDataSourceFormat'] = _FILEDATASOURCEFORMAT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TimestampColumnProperties = _reflection.GeneratedProtocolMessageType('TimestampColumnProperties', (_message.Message,), {
  'DESCRIPTOR' : _TIMESTAMPCOLUMNPROPERTIES,
  '__module__' : 'tecton_proto.data.batch_data_source_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.TimestampColumnProperties)
  })
_sym_db.RegisterMessage(TimestampColumnProperties)

BatchDataSource = _reflection.GeneratedProtocolMessageType('BatchDataSource', (_message.Message,), {
  'DESCRIPTOR' : _BATCHDATASOURCE,
  '__module__' : 'tecton_proto.data.batch_data_source_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.BatchDataSource)
  })
_sym_db.RegisterMessage(BatchDataSource)

FileDataSource = _reflection.GeneratedProtocolMessageType('FileDataSource', (_message.Message,), {
  'DESCRIPTOR' : _FILEDATASOURCE,
  '__module__' : 'tecton_proto.data.batch_data_source_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.FileDataSource)
  })
_sym_db.RegisterMessage(FileDataSource)

DatetimePartitionColumn = _reflection.GeneratedProtocolMessageType('DatetimePartitionColumn', (_message.Message,), {
  'DESCRIPTOR' : _DATETIMEPARTITIONCOLUMN,
  '__module__' : 'tecton_proto.data.batch_data_source_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.DatetimePartitionColumn)
  })
_sym_db.RegisterMessage(DatetimePartitionColumn)

RedshiftDataSource = _reflection.GeneratedProtocolMessageType('RedshiftDataSource', (_message.Message,), {
  'DESCRIPTOR' : _REDSHIFTDATASOURCE,
  '__module__' : 'tecton_proto.data.batch_data_source_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.RedshiftDataSource)
  })
_sym_db.RegisterMessage(RedshiftDataSource)

SnowflakeDataSource = _reflection.GeneratedProtocolMessageType('SnowflakeDataSource', (_message.Message,), {
  'DESCRIPTOR' : _SNOWFLAKEDATASOURCE,
  '__module__' : 'tecton_proto.data.batch_data_source_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.SnowflakeDataSource)
  })
_sym_db.RegisterMessage(SnowflakeDataSource)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



class TectonValidationError(ValueError):
    """
    Exception that indicates a problem in validating user inputs against
    the data in the system. Typically recoverable by the user.
    """


class TectonInternalError(RuntimeError):
    """
    Exception that indicates an unexpected error within Tecton.
    Can be persistent or transient. Recovery typically requires involving
    Tecton support.
    """



import random
import string
from typing import *

import pandas
import pandas as pd
import pendulum
from pyspark.sql import DataFrame
from pyspark.sql import SparkSession
from pyspark.sql.types import ArrayType
from pyspark.sql.types import BooleanType
from pyspark.sql.types import DoubleType
from pyspark.sql.types import FloatType
from pyspark.sql.types import LongType
from pyspark.sql.types import StringType
from pyspark.sql.types import StructField
from pyspark.sql.types import StructType

from tecton_proto.args.new_transformation_pb2 import NewTransformationArgs as TransformationArgs
from tecton_proto.args.new_transformation_pb2 import TransformationMode
from tecton_proto.args.pipeline_pb2 import DataSourceNode
from tecton_proto.args.pipeline_pb2 import Input as InputProto
from tecton_proto.args.pipeline_pb2 import Pipeline
from tecton_proto.args.pipeline_pb2 import PipelineNode
from tecton_proto.args.pipeline_pb2 import RequestContext as RequestContextProto
from tecton_proto.args.pipeline_pb2 import TransformationNode
from tecton_proto.data.new_transformation_pb2 import NewTransformation as Transformation
from tecton_proto.data.virtual_data_source_pb2 import VirtualDataSource
from tecton_spark import data_source_helper
from tecton_spark import function_serialization
from tecton_spark import time_utils
from tecton_spark.id_helper import IdHelper
from tecton_spark.materialization_context import BaseMaterializationContext
from tecton_spark.materialization_context import BoundMaterializationContext
from tecton_spark.pipeline_common import constant_node_to_value
from tecton_spark.pipeline_common import CONSTANT_TYPE
from tecton_spark.pipeline_common import get_keyword_inputs
from tecton_spark.pipeline_common import get_transformation_name
from tecton_spark.pipeline_common import positional_inputs
from tecton_spark.pipeline_common import transformation_type_checker
from tecton_spark.spark_schema_wrapper import SparkSchemaWrapper

MAX_INT64 = (2**63) - 1


class RequestDataSource:
    """
    Declare a ``RequestDataSource``, for using request-time data in an ``OnDemandFeatureView``.
    """

    def __init__(
        self,
        request_schema: StructType,
    ):
        """
        Creates a new RequestDataSource

        :param request_schema: PySpark schema for the RequestDataSource inputs.

        Example of a RequestDataSource declaration:

        .. code-block:: python

            from tecton import RequestDataSource
            from pyspark.sql.types import DoubleType, StructType, StructField

            request_schema = StructType([StructField('amount', DoubleType())])
            transaction_request = RequestDataSource(request_schema=request_schema)
        """
        self.request_schema = request_schema
        self.id = IdHelper.from_string(IdHelper.generate_string_id())


# Pandas Pipeline (ODFV)
# input_df (spark df) is the spine passed in by the user (including request context),
# and it has been augmented with dependent fv fields in of the form "_udf_internal_{input_name}_{odfv_id}".
# The spark dataframe we return will be everything from the spine, with the on-demand features added
#
# NB: If the user defines their transformation to produce extra columns (besides what's specified in output_schema) they will be ignored
# And if they are missing columns they will fail in this function during runtime.
def dataframe_with_input(
    spark: SparkSession,
    pipeline: Pipeline,
    # This should have data from all inputs
    input_df: DataFrame,
    output_schema: StructType,
    transformations,
    name: str,
    fv_id: str,
    namespace: Optional[str],
) -> DataFrame:
    # pass in only the non-internal fields and udf-internal fields corresponding to this particular odfv
    udf_args = [f"{c.name}" for c in input_df.schema if ("_udf_internal" not in c.name or fv_id in c.name)]
    udf_arg_idx_map = {}
    for idx in range(len(udf_args)):
        udf_arg_idx_map[udf_args[idx]] = idx
    builder = _ODFVPipelineBuilder(
        name=name,
        fv_id=fv_id,
        pipeline=pipeline,
        transformations=transformations,
        udf_arg_idx_map=udf_arg_idx_map,
        output_schema=output_schema,
    )

    from pyspark.sql.functions import col, udf, pandas_udf, from_json

    input_columns = [f"`{c.name}`" for c in input_df.schema]
    odfv_output_object = "_odfv_output"
    if namespace is None:
        namespace = name
    output_columns = [col(f"{odfv_output_object}.{c.name}").alias(f"{namespace}.{c.name}") for c in output_schema]
    if builder.mode == "python":
        _odfv_udf = udf(builder.py_wrapper, output_schema)
        return input_df.select(
            *input_columns, _odfv_udf(*[f"`{c}`" for c in udf_args]).alias(odfv_output_object)
        ).select(*input_columns, *output_columns)
    else:
        assert builder.mode == "pandas"
        # Note: from_json will return null in the case of an unparseable string.
        _odfv_udf = pandas_udf(builder.pandas_udf_wrapper, StringType())
        return input_df.select(
            *input_columns, from_json(_odfv_udf(*[f"`{c}`" for c in udf_args]), output_schema).alias(odfv_output_object)
        ).select(*input_columns, *output_columns)


# TODO: if the run api should support some type of mock inputs other than dicts, then we'd need to modify this
# For now, the same pipeline evaluation works for both.
def run_mock_odfv_pipeline(
    pipeline: Pipeline,
    transformations: List[Transformation],
    name: str,
    mock_inputs: Dict[str, Union[Dict[str, Any], pandas.DataFrame]],
) -> Union[Dict[str, Any], pd.DataFrame]:
    builder = _ODFVPipelineBuilder(
        name=name,
        pipeline=pipeline,
        transformations=transformations,
        udf_arg_idx_map={},
        output_schema=None,
        mock_inputs=mock_inputs,
    )
    return builder._udf_node_to_value(pipeline.root)


def pipeline_to_dataframe(
    spark: SparkSession,
    pipeline: Pipeline,
    consume_streaming_data_sources: bool,
    data_sources: List[VirtualDataSource],
    transformations: List[Transformation],
    feature_time_limits: Optional[pendulum.Period] = None,
    schedule_interval: Optional[pendulum.Duration] = None,
    mock_inputs: Optional[Dict[str, DataFrame]] = None,
) -> DataFrame:
    return _PipelineBuilder(
        spark,
        pipeline,
        consume_streaming_data_sources,
        data_sources,
        transformations,
        feature_time_limits,
        schedule_interval=schedule_interval,
        mock_inputs=mock_inputs,
    ).get_dataframe()


# (validly) assumes we have at most a single request data source in the pipeline
def find_request_context(node: PipelineNode) -> Optional[RequestContextProto]:
    if node.HasField("request_data_source_node"):
        return node.request_data_source_node.request_context
    elif node.HasField("transformation_node"):
        for child in node.transformation_node.inputs:
            rc = find_request_context(child.node)
            if rc is not None:
                return rc
    return None


def get_all_input_keys(node: PipelineNode) -> Set[str]:
    names_set = set()
    _get_all_input_keys_helper(node, names_set)
    return names_set


def _get_all_input_keys_helper(node: PipelineNode, names_set: Set[str]):
    if node.HasField("request_data_source_node"):
        names_set.add(node.request_data_source_node.input_name)
    elif node.HasField("data_source_node"):
        names_set.add(node.data_source_node.input_name)
    elif node.HasField("feature_view_node"):
        names_set.add(node.feature_view_node.input_name)
    elif node.HasField("transformation_node"):
        for child in node.transformation_node.inputs:
            _get_all_input_keys_helper(child.node, names_set)
    return names_set


def get_all_input_ds_id_map(node: PipelineNode) -> Dict[str, str]:
    names_dict = dict()
    _get_all_input_ds_id_map_helper(node, names_dict)
    return names_dict


def _get_all_input_ds_id_map_helper(node: PipelineNode, names_dict: Dict[str, str]):
    if node.HasField("data_source_node"):
        names_dict[node.data_source_node.input_name] = IdHelper.to_string(node.data_source_node.virtual_data_source_id)
    elif node.HasField("transformation_node"):
        for child in node.transformation_node.inputs:
            _get_all_input_ds_id_map_helper(child.node, names_dict)
    return names_dict


# Constructs empty data frames matching schema of DS inputs for the purpose of
# schema-validating the transformation pipeline.
def populate_empty_mock_inputs(
    node: PipelineNode,
    ds_map: Dict[str, VirtualDataSource],
    spark: SparkSession,
):
    empty_mock_inputs = {}
    _populate_empty_mock_inputs_helper(node, empty_mock_inputs, ds_map, spark)
    return empty_mock_inputs


def _populate_empty_mock_inputs_helper(
    node: PipelineNode,
    empty_mock_inputs: Dict[str, DataFrame],
    ds_map: Dict[str, VirtualDataSource],
    spark: SparkSession,
):
    if node.HasField("data_source_node"):
        ds_id = IdHelper.to_string(node.data_source_node.virtual_data_source_id)
        ds_schema = ds_map[ds_id].batch_data_source.spark_schema
        empty_mock_inputs[node.data_source_node.input_name] = spark.createDataFrame(
            [], SparkSchemaWrapper.from_proto(ds_schema).unwrap()
        )
    elif node.HasField("transformation_node"):
        for child in node.transformation_node.inputs:
            _populate_empty_mock_inputs_helper(child.node, empty_mock_inputs, ds_map, spark)


# This class is for Spark pipelines
class _PipelineBuilder:
    # The value of internal nodes in the tree
    _VALUE_TYPE = Union[DataFrame, CONSTANT_TYPE, BaseMaterializationContext]

    def __init__(
        self,
        spark: SparkSession,
        pipeline: Pipeline,
        consume_streaming_data_sources: bool,
        data_sources: List[VirtualDataSource],
        # we only use mode and name from these
        transformations: Union[List[Transformation], List[TransformationArgs]],
        feature_time_limits: Optional[pendulum.Period],
        schedule_interval: Optional[pendulum.Duration] = None,
        mock_inputs: Optional[Dict[str, DataFrame]] = None,
    ):
        self._spark = spark
        self._pipeline = pipeline
        self._consume_streaming_data_sources = consume_streaming_data_sources
        self._feature_time_limits = feature_time_limits
        self._id_to_ds = {IdHelper.to_string(ds.virtual_data_source_id): ds for ds in data_sources}
        self._id_to_transformation = {IdHelper.to_string(t.transformation_id): t for t in transformations}

        self._registered_temp_view_names: List[str] = []
        self._schedule_interval = schedule_interval

        self._mock_inputs = mock_inputs

    def get_dataframe(self) -> DataFrame:
        df = self._node_to_value(self._pipeline.root)
        # Cleanup any temporary tables created during the process
        for temp_name in self._registered_temp_view_names:
            self._spark.sql(f"DROP TABLE {temp_name}")
        assert isinstance(df, DataFrame)
        return df

    def _node_to_value(self, pipeline_node: PipelineNode) -> _VALUE_TYPE:
        if pipeline_node.HasField("transformation_node"):
            return self._transformation_node_to_dataframe(pipeline_node.transformation_node)
        elif pipeline_node.HasField("data_source_node"):
            data_source_node = pipeline_node.data_source_node
            if self._mock_inputs is not None and pipeline_node.data_source_node.input_name in self._mock_inputs:
                ds_id = IdHelper.to_string(data_source_node.virtual_data_source_id)
                raw_data_time_limits = self._get_time_window_from_data_source_node(data_source_node)
                if ds_id not in self._id_to_ds or raw_data_time_limits is None:
                    return self._mock_inputs[pipeline_node.data_source_node.input_name]
                return data_source_helper.apply_partition_and_timestamp_filter(
                    df=self._mock_inputs[pipeline_node.data_source_node.input_name],
                    data_source=self._id_to_ds[ds_id].batch_data_source,
                    raw_data_time_limits=self._get_time_window_from_data_source_node(data_source_node),
                )
            else:
                return self._data_source_node_to_dataframe(pipeline_node.data_source_node)
        elif pipeline_node.HasField("constant_node"):
            return constant_node_to_value(pipeline_node.constant_node)
        elif pipeline_node.HasField("materialization_context_node"):
            if self._feature_time_limits is not None:
                feature_start_time = self._feature_time_limits.start
                feature_end_time = self._feature_time_limits.end
                batch_schedule = self._schedule_interval
            else:
                feature_start_time = pendulum.from_timestamp(0, pendulum.tz.UTC)
                feature_end_time = pendulum.datetime(2100, 1, 1)
                batch_schedule = self._schedule_interval or pendulum.duration()
            return BoundMaterializationContext._create_internal(feature_start_time, feature_end_time, batch_schedule)
        elif pipeline_node.HasField("request_data_source_node"):
            raise ValueError("RequestDataSource is not supported in Spark pipelines")
        elif pipeline_node.HasField("feature_view_node"):
            raise ValueError("Dependent FeatureViews are not supported in Spark pipelines")
        else:
            raise KeyError(f"Unknown PipelineNode type: {pipeline_node}")

    def _transformation_node_to_dataframe(self, transformation_node: TransformationNode) -> DataFrame:
        """Recursively translates inputs to values and then passes them to the transformation."""
        args: List[Union[DataFrame, str, int, float, bool]] = []
        kwargs = {}
        for transformation_input in transformation_node.inputs:
            node_value = self._node_to_value(transformation_input.node)
            if transformation_input.HasField("arg_index"):
                assert len(args) == transformation_input.arg_index
                args.append(node_value)
            elif transformation_input.HasField("arg_name"):
                kwargs[transformation_input.arg_name] = node_value
            else:
                raise KeyError(f"Unknown argument type for Input node: {transformation_input}")

        return self._apply_transformation_function(transformation_node, args, kwargs)

    def _apply_transformation_function(
        self, transformation_node, args, kwargs
    ) -> Union[Dict[str, Any], pd.DataFrame, DataFrame]:
        """For the given transformation node, returns the corresponding DataFrame transformation.

        If needed, resulted function is wrapped with a function that translates mode-specific input/output types to DataFrames.
        """
        transformation = self._id_to_transformation[IdHelper.to_string(transformation_node.transformation_id)]
        user_function = function_serialization.from_proto(transformation.user_function)
        transformation_name = get_transformation_name(transformation)

        if transformation.transformation_mode == TransformationMode.TRANSFORMATION_MODE_PYSPARK:
            res = user_function(*args, **kwargs)
            transformation_type_checker(transformation_name, res, "pyspark", self._possible_modes())
            return res
        elif transformation.transformation_mode == TransformationMode.TRANSFORMATION_MODE_SPARK_SQL:
            # type checking happens inside this function
            return self._wrap_sql_function(transformation_node, user_function)(*args, **kwargs)
        elif transformation.transformation_mode == TransformationMode.TRANSFORMATION_MODE_PANDAS:
            res = user_function(*args, **kwargs)
            transformation_type_checker(transformation_name, res, "pandas", self._possible_modes())
            return res
        elif transformation.transformation_mode == TransformationMode.TRANSFORMATION_MODE_PYTHON:
            res = user_function(*args, **kwargs)
            # Only restrict types on the root node of python-mode transforms
            if transformation_node == self._pipeline.root:
                transformation_type_checker(transformation_name, res, "python", self._possible_modes())
            return res
        else:
            raise KeyError(f"Unknown transformation mode: {transformation.transformation_mode}")

    def _wrap_sql_function(
        self, transformation_node: TransformationNode, user_function: Callable[..., str]
    ) -> Callable[..., DataFrame]:
        def wrapped(*args, **kwargs):
            wrapped_args = []
            for arg, node_input in zip(args, positional_inputs(transformation_node)):
                wrapped_args.append(self._wrap_node_inputvalue(node_input, arg))
            keyword_inputs = get_keyword_inputs(transformation_node)
            wrapped_kwargs = {}
            for k, v in kwargs.items():
                node_input = keyword_inputs[k]
                wrapped_kwargs[k] = self._wrap_node_inputvalue(node_input, v)
            sql_string = user_function(*wrapped_args, **wrapped_kwargs)
            transformation_name = get_transformation_name(
                self._id_to_transformation[IdHelper.to_string(transformation_node.transformation_id)]
            )
            transformation_type_checker(transformation_name, sql_string, "spark_sql", self._possible_modes())
            return self._spark.sql(sql_string)

        return wrapped

    def _wrap_node_inputvalue(
        self, node_input, value: _VALUE_TYPE
    ) -> Optional[Union[InputProto, str, int, float, bool]]:
        if node_input.node.HasField("constant_node"):
            assert (
                isinstance(value, str)
                or isinstance(value, int)
                or isinstance(value, float)
                or isinstance(value, bool)
                or value is None
            )
            return value
        elif node_input.node.HasField("materialization_context_node"):
            assert isinstance(value, BoundMaterializationContext)
            return value
        else:
            assert isinstance(value, DataFrame)
            return self._register_temp_table(self._node_name(node_input.node), value)

    def _node_name(self, node) -> str:
        """Returns a human-readable name for the node."""
        if node.HasField("transformation_node"):
            name = get_transformation_name(
                self._id_to_transformation[IdHelper.to_string(node.transformation_node.transformation_id)]
            )
            return f"transformation_{name}_output"
        elif node.HasField("data_source_node"):
            if node.data_source_node.HasField("input_name"):
                return node.data_source_node.input_name
            # TODO(TEC-5076): remove this legacy code, since input_name will always be set
            name = self._id_to_ds[IdHelper.to_string(node.data_source_node.virtual_data_source_id)].fco_metadata.name
            return f"data_source_{name}_output"
        else:
            raise Exception(f"Expected transformation or data source node: {node}")

    def _register_temp_table(self, name: str, df: DataFrame) -> str:
        """Registers a Dataframe as a temp table and returns its name."""
        unique_name = name + self._random_suffix()
        self._registered_temp_view_names.append(unique_name)
        df.createOrReplaceTempView(unique_name)
        return unique_name

    def _random_suffix(self) -> str:
        return "".join(random.choice(string.ascii_letters) for i in range(6))

    def _data_source_node_to_dataframe(self, data_source_node: DataSourceNode) -> DataFrame:
        """Creates a DataFrame from a DS and time parameters."""
        ds = self._id_to_ds[IdHelper.to_string(data_source_node.virtual_data_source_id)]
        time_window = self._get_time_window_from_data_source_node(data_source_node)
        return data_source_helper.get_ds_dataframe(
            self._spark,
            ds,
            self._consume_streaming_data_sources,
            raw_data_time_limits=time_window,
        )

    def _get_time_window_from_data_source_node(self, data_source_node: DataSourceNode) -> Optional[pendulum.Period]:
        raw_data_limits = self._feature_time_limits
        if data_source_node.HasField("window") and raw_data_limits:
            new_start = raw_data_limits.start - time_utils.proto_to_duration(data_source_node.window)
            if self._schedule_interval:
                new_start = new_start + self._schedule_interval
            raw_data_limits = pendulum.Period(new_start, raw_data_limits.end)
        elif data_source_node.window_unbounded_preceding and raw_data_limits:
            raw_data_limits = pendulum.Period(pendulum.datetime(1970, 1, 1), raw_data_limits.end)
        elif data_source_node.window_unbounded:
            raw_data_limits = None
        return raw_data_limits

    def _possible_modes(self):
        # note that pipeline is included since this is meant to be a user hint, and it's
        # theoretically possible a pipeline wound up deeper than expected
        return ["pyspark", "spark_sql", "pipeline"]


# For Pandas-mode:
# We need to take the call a udf constructed from the pipeline that will generate the on-demand features.
# A pandas udf takes as inputs (pd.Series...) and outputs pd.Series.
# However, the user-defined transforms take as input pd.DataFrame and output pd.DataFrame.
# We use _ODFVPipelineBuilder to construct a udf wrapper function that translates the inputs and outputs and
# performs some type checking.
# The general idea is that each Node of the pipeline evaluates to a pandas.DataFrame.
# This is what we want since the user-defined transforms take pandas.DataFrame as inputs both from RequestDataSourceNode or FeatureViewNode.
# pandas_udf_wrapper then typechecks and translates the final pandas.DataFrame into a jsonized pandas.Series to match what spark expects.
#
# For Python-mode, we can use a simpler wrapper function for the udf because we don't do any spark<->pandas type conversions.
class _ODFVPipelineBuilder(_PipelineBuilder):
    def __init__(
        self,
        name: str,
        pipeline: Pipeline,
        transformations: List[Transformation],
        # maps input + feature name to arg index that udf function wrapper will be called with.
        # this is needed because we need to know which pd.Series that are inputs to this function correspond to the desired request context fields or dependent fv features that the customer-defined udf uses.
        udf_arg_idx_map: Dict[str, int],
        output_schema: StructType,
        # the id of this OnDemandFeatureView; only required to be set when reading from source data
        fv_id: Optional[str] = None,
        mock_inputs: Optional[Dict[str, Union[Dict[str, Any], pandas.DataFrame]]] = None,
    ):
        self._pipeline = pipeline
        self._name = name
        self._fv_id = fv_id
        self.udf_arg_idx_map = udf_arg_idx_map
        self._id_to_transformation = {IdHelper.to_string(t.transformation_id): t for t in transformations}
        self._output_schema = output_schema
        self._mock_inputs = mock_inputs
        # In Spark, the UDF cannot reference a proto enum, so instead save mode as a string
        self.mode = (
            "python"
            if self._id_to_transformation[
                IdHelper.to_string(self._pipeline.root.transformation_node.transformation_id)
            ].transformation_mode
            == TransformationMode.TRANSFORMATION_MODE_PYTHON
            else "pandas"
        )

    # FOR PYTHON
    def py_wrapper(self, *args):
        assert self.mode == "python"
        self._udf_args: List = args
        res = self._udf_node_to_value(self._pipeline.root)
        return res

    # FOR PANDAS
    def pandas_udf_wrapper(self, *args):
        assert self.mode == "pandas"
        SPARK_TO_PANDAS_TYPES = {
            LongType(): "int64",
            DoubleType(): "float64",
            StringType(): "object",
            BooleanType(): "bool",
            ArrayType(LongType()): "object",
            ArrayType(FloatType()): "object",
            ArrayType(DoubleType()): "object",
            ArrayType(StringType()): "object",
        }

        # self.udf_arg_idx_map tells us which of these pd.Series correspond to a given RequestDataSource or FeatureView input
        self._udf_args: List[pd.Series] = args

        import pandas as pd
        import json
        import numpy as np

        output_df = self._udf_node_to_value(self._pipeline.root)

        assert (
            type(output_df) == pd.DataFrame
        ), f"Transformer returns {str(output_df)}, but must return a pandas.DataFrame instead."

        def _convert_output_array(arr: Union[List, np.array], field: StructField) -> List:
            # UDF array outputs may be either lists or numpy arrays. Numeric lists/arrays may contain a mix of NaN and
            # None (in some cases). To have a consistent behavior, we cast all NaNs to None.
            if field.dataType == ArrayType(LongType()):
                if isinstance(arr, list):
                    arr = np.array(arr)

                if arr.dtype == np.int64:
                    return arr.tolist()
                elif arr.dtype in (np.float64, np.object):
                    # pd.isnull() matches on NaN in numeric arrays and None in object arrays.
                    null_indices = pd.isnull(arr)
                    # Remove NaN/None before casting the array to int64s.
                    int64_arr = np.where(null_indices, 0, arr).astype(np.int64)
                    # Add NaN/None back in as None.
                    int64_and_nulls_arr = np.where(null_indices, None, int64_arr)
                    return int64_and_nulls_arr.tolist()
                else:
                    raise TypeError(
                        f"UDF output array with type {arr.dtype} for {field.name}, which is invalid type for {field.dataType}."
                    )
            elif field.dataType in (ArrayType(FloatType()), ArrayType(DoubleType())):
                if isinstance(arr, list):
                    arr = np.array(arr)

                if arr.dtype in (np.float32, np.object):
                    arr = arr.astype(np.float64)
                elif arr.dtype != np.float64:
                    raise TypeError(
                        f"UDF output array with type {arr.dtype} for {field.name}, which is invalid type for {field.dataType}."
                    )

                # Replace NaNs with None.
                return np.where(np.isnan(arr), None, arr).tolist()
            elif field.dataType == ArrayType(StringType()):
                if isinstance(arr, list):
                    return arr
                else:
                    return arr.tolist()
            else:
                raise TypeError(f"Unexpected array data type: {field.dataType}")

        for field in self._output_schema:
            assert field.name in output_df.columns, (
                f"Expected output schema field '{field.name}' not found in columns of DataFrame returned by "
                f"'{self._name}': [" + ", ".join(output_df.columns) + "]"
            )
            assert SPARK_TO_PANDAS_TYPES[field.dataType] == output_df[field.name].dtype.name, (
                f"DataFrame field '{field.name}' returned from '{self._name}' has type '{output_df[field.name].dtype.name}', "
                f"but is expected to have Pandas dtype '{SPARK_TO_PANDAS_TYPES[field.dataType]}' (corresponding to Spark type {field.dataType.__class__.__name__}). "
                "See https://pandas.pydata.org/pandas-docs/stable/getting_started/basics.html#dtypes for more info on Pandas dtypes."
            )

            # Convert np.arrays to python lists which are JSON serializable by the default serializer. Also handle edge
            # cases around nulls and NaNs.
            if isinstance(field.dataType, ArrayType):
                output_df[field.name] = output_df[field.name].apply(lambda arr: _convert_output_array(arr, field))

        output_strs = []

        # itertuples() is used instead of iterrows() to preserve type safety.
        # See notes in https://pandas.pydata.org/pandas-docs/version/0.17.1/generated/pandas.DataFrame.iterrows.html.
        for row in output_df.itertuples(index=False):
            output_strs.append(json.dumps(row._asdict()))

        return pd.Series(output_strs)

    def _transformation_node_to_online_dataframe(
        self, transformation_node: TransformationNode
    ) -> Union[Dict[str, Any], pd.DataFrame]:
        """Recursively translates inputs to values and then passes them to the transformation."""
        args: List[Union[DataFrame, str, int, float, bool]] = []
        kwargs = {}
        for transformation_input in transformation_node.inputs:
            node_value = self._udf_node_to_value(transformation_input.node)
            if transformation_input.HasField("arg_index"):
                assert len(args) == transformation_input.arg_index
                args.append(node_value)
            elif transformation_input.HasField("arg_name"):
                kwargs[transformation_input.arg_name] = node_value
            else:
                raise KeyError(f"Unknown argument type for Input node: {transformation_input}")

        return self._apply_transformation_function(transformation_node, args, kwargs)

    # evaluate a node in the Pipeline
    def _udf_node_to_value(
        self, pipeline_node: PipelineNode
    ) -> Union[str, int, float, bool, None, Dict[str, Any], pd.DataFrame, DataFrame, pd.Series]:
        if pipeline_node.HasField("constant_node"):
            return constant_node_to_value(pipeline_node.constant_node)
        elif pipeline_node.HasField("feature_view_node"):
            if self._mock_inputs is not None:
                return self._mock_inputs[pipeline_node.feature_view_node.input_name]
            elif self.mode == "python":
                fields_dict = {}
                # The input name of this FeatureViewNode tells us which of the udf_args correspond to the Dict we should generate that the parent TransformationNode expects as an input.
                # It also expects the DataFrame to have its columns named by the feature names.
                for feature in self.udf_arg_idx_map:
                    if not feature.startswith(
                        f"_udf_internal_{pipeline_node.feature_view_node.input_name}_{self._fv_id}"
                    ):
                        continue
                    idx = self.udf_arg_idx_map[feature]
                    fields_dict[feature.split(".")[1]] = self._udf_args[idx]
                return fields_dict
            elif self.mode == "pandas":
                all_series = []
                features = []
                # The input name of this FeatureViewNode tells us which of the udf_args correspond to the pandas.DataFrame we should generate that the parent TransformationNode expects as an input.
                # It also expects the DataFrame to have its columns named by the feature names.
                for feature in self.udf_arg_idx_map:
                    if not feature.startswith(
                        f"_udf_internal_{pipeline_node.feature_view_node.input_name}_{self._fv_id}"
                    ):
                        continue
                    idx = self.udf_arg_idx_map[feature]
                    all_series.append(self._udf_args[idx])
                    features.append(feature.split(".")[1])
                df = pd.concat(all_series, keys=features, axis=1)
                return df
            else:
                raise NotImplementedError("Transform mode {self.mode} is not yet implemented")
        elif pipeline_node.HasField("request_data_source_node"):
            if self._mock_inputs is not None:
                return self._mock_inputs[pipeline_node.request_data_source_node.input_name]
            elif self.mode == "python":
                request_context = pipeline_node.request_data_source_node.request_context
                field_names = [field.name for field in request_context.schema.fields]
                fields_dict = {}
                for input_col in field_names:
                    idx = self.udf_arg_idx_map[input_col]
                    fields_dict[input_col] = self._udf_args[idx]
                return fields_dict
            elif self.mode == "pandas":
                all_series = []
                request_context = pipeline_node.request_data_source_node.request_context
                field_names = [field.name for field in request_context.schema.fields]
                for input_col in field_names:
                    idx = self.udf_arg_idx_map[input_col]
                    all_series.append(self._udf_args[idx])
                df = pd.concat(all_series, keys=field_names, axis=1)
                return df
            else:
                raise NotImplementedError("Transform mode {self.mode} is not yet implemented")
        elif pipeline_node.HasField("transformation_node"):
            return self._transformation_node_to_online_dataframe(pipeline_node.transformation_node)
        elif pipeline_node.HasField("materialization_context_node"):
            raise ValueError("MaterializationContext is unsupported for pandas pipelines")
        else:
            raise NotImplementedError("This is not yet implemented")

    def _possible_modes(self):
        # note that pipeline is included since this is meant to be a user hint, and it's
        # theoretically possible a pipeline wound up deeper than expected
        return ["pandas", "pipeline", "python"]



import re
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Union

import pandas
import pendulum
import sqlparse

from tecton_proto.common import aggregation_function_pb2 as afpb
from tecton_proto.data.feature_view_pb2 import FeatureView as FeatureViewProto
from tecton_proto.data.feature_view_pb2 import NewTemporalAggregate
from tecton_spark import conf
from tecton_spark import time_utils
from tecton_spark.feature_set_config import FeatureSetConfig
from tecton_spark.snowflake.errors import TectonSnowflakeNotImplementedError
from tecton_spark.snowflake.pipeline_helper import pipeline_to_df_with_input
from tecton_spark.snowflake.pipeline_helper import pipeline_to_sql_string
from tecton_spark.snowflake.templates_utils import load_template

TEMP_SPINE_TABLE_NAME = "_TEMP_SPINE_TABLE_FROM_DF"

HISTORICAL_FEATURES_TEMPLATE = None
MATERIALIZATION_TILE_TEMPLATE = None
MATERIALIZATION_TEMPLATE = None
TIME_LIMIT_TEMPLATE = None
FULL_AGGREGATION_TEMPLATE = None
PARTIAL_AGGREGATION_TEMPLATE = None

# TODO(TEC-6204): Last and LastN are not currently supported
#
# Map of proto function type -> set of (output column prefix, snowflake function name)
AGGREGATION_PLANS = {
    afpb.AGGREGATION_FUNCTION_SUM: {("SUM", "SUM")},
    afpb.AGGREGATION_FUNCTION_MIN: {("MIN", "MIN")},
    afpb.AGGREGATION_FUNCTION_MAX: {("MAX", "MAX")},
    afpb.AGGREGATION_FUNCTION_COUNT: {("COUNT", "COUNT")},
    afpb.AGGREGATION_FUNCTION_MEAN: {("COUNT", "COUNT"), ("MEAN", "AVG")},
}


def _load_template():
    # TODO: Do this at module loading time once we sort out including the templates in the public SDK build
    global HISTORICAL_FEATURES_TEMPLATE
    if not HISTORICAL_FEATURES_TEMPLATE:
        HISTORICAL_FEATURES_TEMPLATE = load_template("historical_features.sql.j2")
    global MATERIALIZATION_TILE_TEMPLATE
    if not MATERIALIZATION_TILE_TEMPLATE:
        MATERIALIZATION_TILE_TEMPLATE = load_template("materialization_tile.sql.j2")
    global MATERIALIZATION_TEMPLATE
    if not MATERIALIZATION_TEMPLATE:
        MATERIALIZATION_TEMPLATE = load_template("materialization.sql.j2")
    global TIME_LIMIT_TEMPLATE
    if not TIME_LIMIT_TEMPLATE:
        TIME_LIMIT_TEMPLATE = load_template("time_limit.sql.j2")
    global FULL_AGGREGATION_TEMPLATE
    if not FULL_AGGREGATION_TEMPLATE:
        FULL_AGGREGATION_TEMPLATE = load_template("run_full_aggregation.sql.j2")
    global PARTIAL_AGGREGATION_TEMPLATE
    if not PARTIAL_AGGREGATION_TEMPLATE:
        PARTIAL_AGGREGATION_TEMPLATE = load_template("run_partial_aggregation.sql.j2")


def _format_sql(sql_str: str) -> str:
    return sqlparse.format(sql_str, reindent=True, keyword_case="upper")


@dataclass
class _FeatureSetItemInput:
    """A simplified version of FeatureSetItem which is passed to the SQL template."""

    name: str
    timestamp_key: str
    join_keys: Dict[str, str]
    features: List[str]
    sql: str
    aggregation: Optional[NewTemporalAggregate]
    ttl_seconds: Optional[int]


def _generate_historical_features_sql(
    spine_sql: str,
    feature_set_config: FeatureSetConfig,
    timestamp_key: str,
    from_registered_view: bool,
    include_feature_view_timestamp_columns: bool = False,
    # Whether to use the registered snowflake view for features.
) -> str:
    _load_template()
    if include_feature_view_timestamp_columns:
        # TODO(TEC-7238): Implement this
        raise TectonSnowflakeNotImplementedError(
            "include_feature_view_timestamp_columns is not implemented for Snowflake"
        )
    feature_set_items = feature_set_config._get_feature_definitions_and_join_configs()
    input_items = []
    for item in feature_set_items:
        fd = item.feature_definition
        if fd.is_on_demand and not conf.get_or_none("ALPHA_SNOWFLAKE_SNOWPARK_ENABLED"):
            raise TectonSnowflakeNotImplementedError("On-demand features are only supported with Snowpark enabled")
        feature_view = fd.fv
        # Change the feature view name if it's for internal udf use.
        if item.namespace.startswith("_udf_internal"):
            name = item.namespace.upper()
        else:
            name = fd.name

        if not fd.is_on_demand:
            join_keys = {key: value for key, value in item.join_keys}
            features = [
                col_name
                for col_name in fd.view_schema.column_names()
                if col_name not in (list(join_keys.keys()) + [fd.timestamp_key])
            ]
            if len(fd.online_serving_index.join_keys) != len(fd.join_keys):
                raise TectonSnowflakeNotImplementedError("Wildcard is not supported for Snowflake")
            if from_registered_view:
                source = f"SELECT * FROM {feature_view.snowflake_data.snowflake_view_name}"
            else:
                source = pipeline_to_sql_string(
                    pipeline=fd.pipeline,
                    data_sources=fd.data_sources,
                    transformations=fd.transformations,
                )
            sql_str = TIME_LIMIT_TEMPLATE.render(
                source=source,
                timestamp_key=fd.timestamp_key,
                # Apply time limit from feature_start_time
                start_time=fd.start_timestamp,
            )
            input_items.append(
                _FeatureSetItemInput(
                    name=name,
                    timestamp_key=fd.timestamp_key,
                    join_keys=join_keys,
                    features=features,
                    sql=sql_str,
                    aggregation=(
                        feature_view.temporal_aggregate if feature_view.HasField("temporal_aggregate") else None
                    ),
                    ttl_seconds=(int(fd.serving_ttl.total_seconds()) if fd.is_temporal else None),
                )
            )
    sql_str = HISTORICAL_FEATURES_TEMPLATE.render(
        feature_set_items=input_items,
        spine_timestamp_key=timestamp_key,
        spine_sql=spine_sql,
        include_feature_view_timestamp_columns=include_feature_view_timestamp_columns,
    )

    return sql_str


def get_historical_features(
    spine: Union[pandas.DataFrame, str],
    connection: "snowflake.connector.Connection",
    feature_set_config: FeatureSetConfig,
    timestamp_key: str,
    include_feature_view_timestamp_columns: bool = False,
    # Whether to use the registered snowflake view for features.
    from_registered_view: bool = False,
) -> pandas.DataFrame:
    cur = connection.cursor()
    if isinstance(spine, str):
        spine_sql = spine
    elif isinstance(spine, pandas.DataFrame):
        from snowflake.connector.pandas_tools import write_pandas

        # Get the SQL that would be generated by the create table statement
        create_table_sql = pandas.io.sql.get_schema(spine, TEMP_SPINE_TABLE_NAME)

        # Replace the `CREATE TABLE` with `CREATE OR REPLACE TEMPORARY TABLE`
        create_tmp_table_sql = re.sub("^(CREATE TABLE)?", "CREATE OR REPLACE TEMPORARY TABLE", create_table_sql)
        cur.execute(create_tmp_table_sql)
        write_pandas(conn=connection, df=spine, table_name=TEMP_SPINE_TABLE_NAME)
        spine_sql = f"SELECT * FROM {TEMP_SPINE_TABLE_NAME}"

    sql_str = _generate_historical_features_sql(
        spine_sql=spine_sql,
        feature_set_config=feature_set_config,
        timestamp_key=timestamp_key,
        include_feature_view_timestamp_columns=include_feature_view_timestamp_columns,
        from_registered_view=from_registered_view,
    )
    cur.execute(sql_str)
    return cur.fetch_pandas_all()


def get_historical_features_with_snowpark(
    spine: Union[pandas.DataFrame, str, "snowflake.snowpark.DataFrame"],
    session: "snowflake.snowpark.Session",
    feature_set_config: FeatureSetConfig,
    timestamp_key: str,
    include_feature_view_timestamp_columns: bool = False,
    # Whether to use the registered snowflake view for features.
    from_registered_view: bool = False,
) -> "snowflake.snowpark.DataFrame":
    from snowflake.snowpark import DataFrame

    if isinstance(spine, str):
        spine_sql = spine
    elif isinstance(spine, DataFrame):
        spine.write.save_as_table(TEMP_SPINE_TABLE_NAME, mode="overwrite", create_temp_table=True)
        spine_sql = f"SELECT * FROM {TEMP_SPINE_TABLE_NAME}"
    elif isinstance(spine, pandas.DataFrame):
        session.sql(f"DROP TABLE IF EXISTS {TEMP_SPINE_TABLE_NAME}").collect()
        session.write_pandas(spine, TEMP_SPINE_TABLE_NAME, auto_create_table=True, create_temp_table=True)
        spine_sql = f"SELECT * FROM {TEMP_SPINE_TABLE_NAME}"

    sql_str = _generate_historical_features_sql(
        spine_sql=spine_sql,
        feature_set_config=feature_set_config,
        timestamp_key=timestamp_key,
        include_feature_view_timestamp_columns=include_feature_view_timestamp_columns,
        from_registered_view=from_registered_view,
    )
    output_df = session.sql(sql_str)

    # Apply ODFV to the spine
    for item in feature_set_config._get_feature_definitions_and_join_configs():
        fd = item.feature_definition
        if fd.is_on_demand:
            view_schema = fd.view_schema
            schema_dict = {col_name: view_schema.spark_type(col_name) for col_name in view_schema.column_names()}
            output_df = pipeline_to_df_with_input(
                session=session,
                input_df=output_df,
                pipeline=fd.pipeline,
                transformations=fd.transformations,
                output_schema=schema_dict,
                name=fd.name,
                fv_id=fd.id,
            )
    columns_to_drop = [column for column in output_df.columns if "_UDF_INTERNAL_" in column]
    if len(columns_to_drop) > 0:
        output_df = output_df.drop(*columns_to_drop)
    return output_df


def generate_run_batch_sql(
    feature_view: FeatureViewProto,
    # start is inclusive and end is exclusive
    feature_start_time: Union[pendulum.DateTime, datetime] = None,
    feature_end_time: Union[pendulum.DateTime, datetime] = None,
    aggregation_level: str = "full",
) -> str:
    # Smart default feature_start_time and feature_end_time if unset.
    if feature_end_time is None:
        feature_end_time = pendulum.now()
    if feature_start_time is None:
        schedule_interval = time_utils.proto_to_duration(feature_view.materialization_params.schedule_interval)
        feature_start_time = feature_end_time - schedule_interval
    if feature_view.HasField("temporal_aggregate"):
        if aggregation_level == "full":
            aggregated_tiles_sql_str = get_materialization_sql_str(
                feature_view=feature_view,
                time_limits=pendulum.period(feature_start_time, feature_end_time),
            )
            aggregated_sql_str = FULL_AGGREGATION_TEMPLATE.render(
                source=aggregated_tiles_sql_str,
                join_keys=list(feature_view.join_keys),
                aggregation=feature_view.temporal_aggregate,
                timestamp_key=feature_view.timestamp_key,
                name=feature_view.fco_metadata.name,
            )
            return _format_sql(aggregated_sql_str)
        elif aggregation_level == "partial":
            aggregated_tiles_sql_str = get_materialization_sql_str(
                feature_view=feature_view,
                time_limits=pendulum.period(feature_start_time, feature_end_time),
            )
            # Rename the output columns, and add tile start/end time columns
            partial_aggregated_sql_str = PARTIAL_AGGREGATION_TEMPLATE.render(
                source=aggregated_tiles_sql_str,
                join_keys=list(feature_view.join_keys),
                aggregations=_get_feature_view_aggregations(feature_view),
                slide_interval=feature_view.temporal_aggregate.slide_interval,
                slide_interval_string=feature_view.temporal_aggregate.slide_interval_string,
                timestamp_key=feature_view.timestamp_key,
            )
            return _format_sql(partial_aggregated_sql_str)
        elif aggregation_level == "disabled":
            sql_str = TIME_LIMIT_TEMPLATE.render(
                source=feature_view.snowflake_data.snowflake_view_name,
                timestamp_key=feature_view.timestamp_key,
                start_time=feature_start_time,
                end_time=feature_end_time,
            )
            return _format_sql(sql_str)
        else:
            raise ValueError(f"Unsupported aggregation level: {aggregation_level}")

    else:
        sql_str = get_materialization_sql_str(
            feature_view=feature_view,
            time_limits=pendulum.period(feature_start_time, feature_end_time),
        )
        return _format_sql(sql_str)


def get_materialization_sql_str(
    feature_view: FeatureViewProto,
    # start is inclusive and end is exclusive
    time_limits: pendulum.Period,
    destination: str = None,
    storage_integration: Optional[str] = None,
) -> str:
    _load_template()
    source = feature_view.snowflake_data.snowflake_view_name
    if feature_view.HasField("temporal_aggregate"):
        source = MATERIALIZATION_TILE_TEMPLATE.render(
            source=source,
            join_keys=list(feature_view.join_keys),
            aggregations=_get_feature_view_aggregations(feature_view),
            slide_interval=feature_view.temporal_aggregate.slide_interval,
            timestamp_key=feature_view.timestamp_key,
        )
    return _format_sql(
        MATERIALIZATION_TEMPLATE.render(
            source=source,
            materialization_schema=feature_view.schemas.materialization_schema,
            timestamp_key=feature_view.timestamp_key,
            start_time=time_limits.start,
            end_time=time_limits.end,
            destination=destination,
            storage_integration=storage_integration,
        )
    )


def get_features_sql_str_for_spine(
    feature_set_config: FeatureSetConfig,
    timestamp_key: str,
    spine_sql: str = None,
    include_feature_view_timestamp_columns: bool = False,
) -> str:
    """
    Get a SQL string to fetch features given the spine and feature set.
    spine_sql and spine_table_name cannot be both empty.

    :param feature_set_config: FeatureSetConfig for the features.
    :param timestamp_key: Name of the time column in the spine.
    :param spine_sql: SQL str to get the spine.
    :param include_feature_view_timestamp_columns: (Optional) Include timestamp columns for every individual feature definitions.
    :return: A SQL string that can be used to fetch features.
    """
    _load_template()
    if include_feature_view_timestamp_columns:
        # TODO(TEC-7238): Implement this
        raise TectonSnowflakeNotImplementedError(
            "include_feature_view_timestamp_columns is not implemented for Snowflake"
        )

    feature_set_items = feature_set_config._get_feature_definitions_and_join_configs()
    input_items = []
    for item in feature_set_items:
        fd = item.feature_definition
        feature_view = fd.fv
        join_keys = {key: value for key, value in item.join_keys}
        features = [
            col_name
            for col_name in fd.view_schema.column_names()
            if col_name not in (list(join_keys.keys()) + [fd.timestamp_key])
        ]
        if len(fd.online_serving_index.join_keys) != len(fd.join_keys):
            raise ValueError("SQL string does not support wildcard")
        sql_str = TIME_LIMIT_TEMPLATE.render(
            source=pipeline_to_sql_string(
                pipeline=fd.pipeline,
                data_sources=fd.data_sources,
                transformations=fd.transformations,
            ),
            timestamp_key=fd.timestamp_key,
            # Apply time limit from feature_start_time
            start_time=fd.start_timestamp,
        )
        input_items.append(
            _FeatureSetItemInput(
                name=fd.name,
                timestamp_key=fd.timestamp_key,
                join_keys=join_keys,
                features=features,
                sql=sql_str,
                aggregation=(feature_view.temporal_aggregate if feature_view.HasField("temporal_aggregate") else None),
                ttl_seconds=(int(fd.serving_ttl.total_seconds()) if fd.is_temporal else None),
            )
        )
    return _format_sql(
        HISTORICAL_FEATURES_TEMPLATE.render(
            feature_set_items=input_items,
            spine_timestamp_key=timestamp_key,
            spine_sql=spine_sql,
            include_feature_view_timestamp_columns=include_feature_view_timestamp_columns,
        )
    )


def _get_feature_view_aggregations(feature_view: FeatureViewProto) -> Dict[str, Set[str]]:
    aggregations = defaultdict(set)
    for feature in feature_view.temporal_aggregate.features:
        aggregate_function = AGGREGATION_PLANS[feature.function]
        if not aggregate_function:
            raise TectonSnowflakeNotImplementedError(
                f"Unsupported aggregation function {feature.function} in snowflake pipeline"
            )
        aggregations[feature.input_feature_name].update(aggregate_function)

    # Need to order the functions for deterministic results.
    for key, value in aggregations.items():
        aggregations[key] = sorted(value)

    return aggregations



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/fco.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2
from tecton_proto.data import entity_pb2 as tecton__proto_dot_data_dot_entity__pb2
from tecton_proto.data import virtual_data_source_pb2 as tecton__proto_dot_data_dot_virtual__data__source__pb2
from tecton_proto.data import feature_view_pb2 as tecton__proto_dot_data_dot_feature__view__pb2
from tecton_proto.data import feature_service_pb2 as tecton__proto_dot_data_dot_feature__service__pb2
from tecton_proto.data import new_transformation_pb2 as tecton__proto_dot_data_dot_new__transformation__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/data/fco.proto',
  package='tecton_proto.data',
  syntax='proto2',
  serialized_options=b'\n\017com.tecton.dataP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1btecton_proto/data/fco.proto\x12\x11tecton_proto.data\x1a\x1ctecton_proto/common/id.proto\x1a\x1etecton_proto/data/entity.proto\x1a+tecton_proto/data/virtual_data_source.proto\x1a$tecton_proto/data/feature_view.proto\x1a\'tecton_proto/data/feature_service.proto\x1a*tecton_proto/data/new_transformation.proto\"\x8f\x03\n\x03\x46\x63o\x12V\n\x13virtual_data_source\x18\x01 \x01(\x0b\x32$.tecton_proto.data.VirtualDataSourceH\x00R\x11virtualDataSource\x12\x33\n\x06\x65ntity\x18\x02 \x01(\x0b\x32\x19.tecton_proto.data.EntityH\x00R\x06\x65ntity\x12\x43\n\x0c\x66\x65\x61ture_view\x18\x06 \x01(\x0b\x32\x1e.tecton_proto.data.FeatureViewH\x00R\x0b\x66\x65\x61tureView\x12L\n\x0f\x66\x65\x61ture_service\x18\x04 \x01(\x0b\x32!.tecton_proto.data.FeatureServiceH\x00R\x0e\x66\x65\x61tureService\x12U\n\x12new_transformation\x18\x07 \x01(\x0b\x32$.tecton_proto.data.NewTransformationH\x00R\x11newTransformationB\x05\n\x03\x66\x63oJ\x04\x08\x03\x10\x04J\x04\x08\x05\x10\x06\"n\n\x0c\x46\x63oContainer\x12\x32\n\x08root_ids\x18\x01 \x03(\x0b\x32\x17.tecton_proto.common.IdR\x07rootIds\x12*\n\x04\x66\x63os\x18\x02 \x03(\x0b\x32\x16.tecton_proto.data.FcoR\x04\x66\x63osB\x13\n\x0f\x63om.tecton.dataP\x01'
  ,
  dependencies=[tecton__proto_dot_common_dot_id__pb2.DESCRIPTOR,tecton__proto_dot_data_dot_entity__pb2.DESCRIPTOR,tecton__proto_dot_data_dot_virtual__data__source__pb2.DESCRIPTOR,tecton__proto_dot_data_dot_feature__view__pb2.DESCRIPTOR,tecton__proto_dot_data_dot_feature__service__pb2.DESCRIPTOR,tecton__proto_dot_data_dot_new__transformation__pb2.DESCRIPTOR,])




_FCO = _descriptor.Descriptor(
  name='Fco',
  full_name='tecton_proto.data.Fco',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='virtual_data_source', full_name='tecton_proto.data.Fco.virtual_data_source', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='virtualDataSource', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='entity', full_name='tecton_proto.data.Fco.entity', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='entity', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_view', full_name='tecton_proto.data.Fco.feature_view', index=2,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='featureView', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_service', full_name='tecton_proto.data.Fco.feature_service', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='featureService', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='new_transformation', full_name='tecton_proto.data.Fco.new_transformation', index=4,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='newTransformation', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='fco', full_name='tecton_proto.data.Fco.fco',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=281,
  serialized_end=680,
)


_FCOCONTAINER = _descriptor.Descriptor(
  name='FcoContainer',
  full_name='tecton_proto.data.FcoContainer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='root_ids', full_name='tecton_proto.data.FcoContainer.root_ids', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='rootIds', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fcos', full_name='tecton_proto.data.FcoContainer.fcos', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='fcos', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=682,
  serialized_end=792,
)

_FCO.fields_by_name['virtual_data_source'].message_type = tecton__proto_dot_data_dot_virtual__data__source__pb2._VIRTUALDATASOURCE
_FCO.fields_by_name['entity'].message_type = tecton__proto_dot_data_dot_entity__pb2._ENTITY
_FCO.fields_by_name['feature_view'].message_type = tecton__proto_dot_data_dot_feature__view__pb2._FEATUREVIEW
_FCO.fields_by_name['feature_service'].message_type = tecton__proto_dot_data_dot_feature__service__pb2._FEATURESERVICE
_FCO.fields_by_name['new_transformation'].message_type = tecton__proto_dot_data_dot_new__transformation__pb2._NEWTRANSFORMATION
_FCO.oneofs_by_name['fco'].fields.append(
  _FCO.fields_by_name['virtual_data_source'])
_FCO.fields_by_name['virtual_data_source'].containing_oneof = _FCO.oneofs_by_name['fco']
_FCO.oneofs_by_name['fco'].fields.append(
  _FCO.fields_by_name['entity'])
_FCO.fields_by_name['entity'].containing_oneof = _FCO.oneofs_by_name['fco']
_FCO.oneofs_by_name['fco'].fields.append(
  _FCO.fields_by_name['feature_view'])
_FCO.fields_by_name['feature_view'].containing_oneof = _FCO.oneofs_by_name['fco']
_FCO.oneofs_by_name['fco'].fields.append(
  _FCO.fields_by_name['feature_service'])
_FCO.fields_by_name['feature_service'].containing_oneof = _FCO.oneofs_by_name['fco']
_FCO.oneofs_by_name['fco'].fields.append(
  _FCO.fields_by_name['new_transformation'])
_FCO.fields_by_name['new_transformation'].containing_oneof = _FCO.oneofs_by_name['fco']
_FCOCONTAINER.fields_by_name['root_ids'].message_type = tecton__proto_dot_common_dot_id__pb2._ID
_FCOCONTAINER.fields_by_name['fcos'].message_type = _FCO
DESCRIPTOR.message_types_by_name['Fco'] = _FCO
DESCRIPTOR.message_types_by_name['FcoContainer'] = _FCOCONTAINER
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Fco = _reflection.GeneratedProtocolMessageType('Fco', (_message.Message,), {
  'DESCRIPTOR' : _FCO,
  '__module__' : 'tecton_proto.data.fco_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.Fco)
  })
_sym_db.RegisterMessage(Fco)

FcoContainer = _reflection.GeneratedProtocolMessageType('FcoContainer', (_message.Message,), {
  'DESCRIPTOR' : _FCOCONTAINER,
  '__module__' : 'tecton_proto.data.fco_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.FcoContainer)
  })
_sym_db.RegisterMessage(FcoContainer)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/tecton_api_key.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/data/tecton_api_key.proto',
  package='tecton_proto.data',
  syntax='proto2',
  serialized_options=b'\n\017com.tecton.dataP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n&tecton_proto/data/tecton_api_key.proto\x12\x11tecton_proto.data\x1a\x1ctecton_proto/common/id.proto\"\xf1\x01\n\x0cTectonApiKey\x12\'\n\x02id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x02id\x12\x1d\n\nhashed_key\x18\x02 \x01(\tR\thashedKey\x12 \n\x0b\x64\x65scription\x18\x03 \x01(\tR\x0b\x64\x65scription\x12\x1a\n\x08\x61rchived\x18\x04 \x01(\x08R\x08\x61rchived\x12\x1d\n\ncreated_by\x18\x05 \x01(\tR\tcreatedBy\x12!\n\x0cobscured_key\x18\x06 \x01(\tR\x0bobscuredKey\x12\x19\n\x08is_admin\x18\x07 \x01(\x08R\x07isAdminB\x13\n\x0f\x63om.tecton.dataP\x01'
  ,
  dependencies=[tecton__proto_dot_common_dot_id__pb2.DESCRIPTOR,])




_TECTONAPIKEY = _descriptor.Descriptor(
  name='TectonApiKey',
  full_name='tecton_proto.data.TectonApiKey',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='tecton_proto.data.TectonApiKey.id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='id', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='hashed_key', full_name='tecton_proto.data.TectonApiKey.hashed_key', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='hashedKey', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='tecton_proto.data.TectonApiKey.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='description', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='archived', full_name='tecton_proto.data.TectonApiKey.archived', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='archived', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='created_by', full_name='tecton_proto.data.TectonApiKey.created_by', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='createdBy', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='obscured_key', full_name='tecton_proto.data.TectonApiKey.obscured_key', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='obscuredKey', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_admin', full_name='tecton_proto.data.TectonApiKey.is_admin', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='isAdmin', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=92,
  serialized_end=333,
)

_TECTONAPIKEY.fields_by_name['id'].message_type = tecton__proto_dot_common_dot_id__pb2._ID
DESCRIPTOR.message_types_by_name['TectonApiKey'] = _TECTONAPIKEY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TectonApiKey = _reflection.GeneratedProtocolMessageType('TectonApiKey', (_message.Message,), {
  'DESCRIPTOR' : _TECTONAPIKEY,
  '__module__' : 'tecton_proto.data.tecton_api_key_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.TectonApiKey)
  })
_sym_db.RegisterMessage(TectonApiKey)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



from datetime import datetime
from typing import List
from typing import Mapping
from typing import Optional
from typing import Union

import pandas as pd
import pendulum
import pyspark
from pyspark.sql import functions

from tecton._internals import data_frame_helper
from tecton._internals import errors as internal_errors
from tecton._internals.feature_views import aggregations
from tecton._internals.utils import is_bfc_mode_single
from tecton._internals.utils import is_live_workspace
from tecton.interactive.data_frame import DataFrame
from tecton.tecton_context import TectonContext
from tecton.tecton_errors import TectonValidationError
from tecton_proto.args.pipeline_pb2 import PipelineNode
from tecton_spark.fco_container import FcoContainer
from tecton_spark.feature_definition_wrapper import FeatureDefinitionWrapper as FeatureDefinition
from tecton_spark.feature_set_config import FeatureDefinitionAndJoinConfig
from tecton_spark.id_helper import IdHelper
from tecton_spark.materialization_params import MaterializationParams
from tecton_spark.schema_spark_utils import schema_to_spark


def get_features(
    fd: FeatureDefinition,
    entities: Optional[Union[pyspark.sql.dataframe.DataFrame, pd.DataFrame, DataFrame]] = None,
    start_time: Optional[pendulum.DateTime] = None,
    end_time: Optional[pendulum.DateTime] = None,
    from_source: bool = False,
) -> DataFrame:
    if fd.is_on_demand:
        raise internal_errors.FV_NOT_SUPPORTED_GET_HISTORICAL_FEATURES

    if from_source and fd.is_feature_table:
        raise TectonValidationError("FeatureTables are not compatible with from_source=True")

    if from_source and is_bfc_mode_single(fd):
        raise internal_errors.FV_BFC_SINGLE_FROM_SOURCE

    if not from_source and not is_live_workspace(fd.workspace):
        raise internal_errors.FD_GET_MATERIALIZED_FEATURES_FROM_DEVELOPMENT_WORKSPACE(fd.name, fd.workspace)

    if not from_source and not fd.writes_to_offline_store:
        raise internal_errors.FD_GET_FEATURES_MATERIALIZATION_DISABLED(fd.name)

    if start_time is not None and isinstance(start_time, datetime):
        start_time = pendulum.instance(start_time)
    if end_time is not None and isinstance(end_time, datetime):
        end_time = pendulum.instance(end_time)

    if fd.is_temporal_aggregate or fd.is_temporal:
        params = MaterializationParams.from_feature_definition(fd)
        assert params is not None, "Materialization params cannot be None"
        _start = start_time or params.from_timestamp
        # we need to add 1 to most_recent_anchor since we filter end_time exclusively
        _end = end_time or (params.most_recent_anchor(pendulum.now("UTC")) + pendulum.duration(microseconds=1))
    else:
        _start = start_time or pendulum.datetime(1970, 1, 1)
        _end = end_time or pendulum.now("UTC")

    time_range = pendulum.Period(_start, _end)

    tc = TectonContext.get_instance()
    spark = tc._spark

    # Validate that entities only contains Join Key Columns.
    if entities is not None:
        if isinstance(entities, pd.DataFrame):
            entities = spark.createDataFrame(entities)
        if isinstance(entities, DataFrame):
            entities = entities.to_spark()
        assert set(entities.columns).issubset(
            set(fd.join_keys)
        ), f"Entities should only contain columns that can be used as Join Keys: {fd.join_keys}"

    try:
        if fd.is_temporal or fd.is_feature_table:
            df = aggregations.get_all_temporal_ft_features(spark, fd, entities, not from_source, time_range)
        else:
            df = data_frame_helper._get_feature_dataframe_with_limits(
                fd,
                spine=None,
                spine_time_limits=time_range,
                use_materialized_data=not from_source,
                spine_time_key=None,
                validate_time_key=False,
            ).to_spark()
            if entities is not None:
                df = df.join(functions.broadcast(entities.distinct()), entities.columns, how="right")
            columns = fd.join_keys + fd.features + [fd.timestamp_key]
            df = df.select(*columns)
    except pyspark.sql.utils.AnalysisException as e:
        if "Unable to infer schema for Parquet" in e.desc or "doesn't exist" in e.desc:
            if fd.is_feature_table:
                return DataFrame._create(tc._spark.createDataFrame([], schema_to_spark(fd.view_schema)))
            else:
                raise internal_errors.FV_NO_MATERIALIZED_DATA(fd.name)
        raise
    # for FVs, raw data filtering should produce data already in the time range for adhoc
    # if FV output is outside feature time range given raw data, we do not want to filter it so the time validation
    # in materialization_plan._validate_feature_timestamps can fail the query. If we filter here, the UDF
    # may be executed after the filter step, when we want it to catch all features generated outside the time range
    # in ad hoc. It's not 100% but for this it may be good enough
    should_filter_by_time = not from_source
    if should_filter_by_time:
        if _start:
            df = df.filter(df[fd.timestamp_key] > _start)
        if _end:
            df = df.filter(df[fd.timestamp_key] < _end)

    return DataFrame._create(df)


def find_dependent_feature_set_items(
    fco_container: FcoContainer, node: PipelineNode, visited_inputs: Mapping[str, bool], fv_id: str, workspace_name: str
) -> List[FeatureDefinitionAndJoinConfig]:
    if node.HasField("feature_view_node"):
        if node.feature_view_node.input_name in visited_inputs:
            return []
        visited_inputs[node.feature_view_node.input_name] = True

        fv_proto = fco_container.get_by_id(IdHelper.to_string(node.feature_view_node.feature_view_id))
        fd = FeatureDefinition(fv_proto, fco_container)

        join_keys = []
        overrides = {
            colpair.feature_column: colpair.spine_column
            for colpair in node.feature_view_node.feature_view.override_join_keys
        }
        for join_key in fv_proto.join_keys:
            potentially_overriden_key = overrides.get(join_key, join_key)
            join_keys.append((potentially_overriden_key, join_key))
        cfg = FeatureDefinitionAndJoinConfig(
            feature_definition=fd,
            name=fd.name,
            join_keys=join_keys,
            namespace=f"_udf_internal_{node.feature_view_node.input_name}_{fv_id}",
            features=node.feature_view_node.feature_view.features or fd.features,
        )

        return [cfg]
    elif node.HasField("transformation_node"):
        ret: List[FeatureDefinitionAndJoinConfig] = []
        for child in node.transformation_node.inputs:
            ret = ret + find_dependent_feature_set_items(
                fco_container, child.node, visited_inputs, fv_id, workspace_name
            )
        return ret
    return []



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/fv_materialization.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.common import fco_locator_pb2 as tecton__proto_dot_common_dot_fco__locator__pb2
from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/data/fv_materialization.proto',
  package='tecton_proto.data',
  syntax='proto2',
  serialized_options=b'\n\017com.tecton.dataP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n*tecton_proto/data/fv_materialization.proto\x12\x11tecton_proto.data\x1a%tecton_proto/common/fco_locator.proto\x1a\x1ctecton_proto/common/id.proto\" \n\nS3Location\x12\x12\n\x04path\x18\x01 \x01(\tR\x04path\"\xed\x03\n\x11\x46vMaterialization\x12Q\n\x14\x66\x65\x61ture_view_locator\x18\t \x01(\x0b\x32\x1f.tecton_proto.common.FcoLocatorR\x12\x66\x65\x61tureViewLocator\x12>\n\x1bmaterialization_workflow_id\x18\x03 \x01(\x05R\x19materializationWorkflowId\x12[\n\x1amaterialized_data_location\x18\x05 \x01(\x0b\x32\x1d.tecton_proto.data.S3LocationR\x18materializedDataLocation\x12\x61\n\x1dstreaming_checkpoint_location\x18\x06 \x01(\x0b\x32\x1d.tecton_proto.data.S3LocationR\x1bstreamingCheckpointLocation\x12\x32\n\x15has_materialized_data\x18\x07 \x01(\x08R\x13hasMaterializedData\x12?\n\x1c\x66\x65\x61ture_store_format_version\x18\x08 \x01(\x05R\x19\x66\x65\x61tureStoreFormatVersionJ\x04\x08\x01\x10\x02J\x04\x08\x02\x10\x03J\x04\x08\x04\x10\x05\x42\x13\n\x0f\x63om.tecton.dataP\x01'
  ,
  dependencies=[tecton__proto_dot_common_dot_fco__locator__pb2.DESCRIPTOR,tecton__proto_dot_common_dot_id__pb2.DESCRIPTOR,])




_S3LOCATION = _descriptor.Descriptor(
  name='S3Location',
  full_name='tecton_proto.data.S3Location',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='path', full_name='tecton_proto.data.S3Location.path', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='path', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=134,
  serialized_end=166,
)


_FVMATERIALIZATION = _descriptor.Descriptor(
  name='FvMaterialization',
  full_name='tecton_proto.data.FvMaterialization',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='feature_view_locator', full_name='tecton_proto.data.FvMaterialization.feature_view_locator', index=0,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='featureViewLocator', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='materialization_workflow_id', full_name='tecton_proto.data.FvMaterialization.materialization_workflow_id', index=1,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='materializationWorkflowId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='materialized_data_location', full_name='tecton_proto.data.FvMaterialization.materialized_data_location', index=2,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='materializedDataLocation', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='streaming_checkpoint_location', full_name='tecton_proto.data.FvMaterialization.streaming_checkpoint_location', index=3,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='streamingCheckpointLocation', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='has_materialized_data', full_name='tecton_proto.data.FvMaterialization.has_materialized_data', index=4,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='hasMaterializedData', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_store_format_version', full_name='tecton_proto.data.FvMaterialization.feature_store_format_version', index=5,
      number=8, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='featureStoreFormatVersion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=169,
  serialized_end=662,
)

_FVMATERIALIZATION.fields_by_name['feature_view_locator'].message_type = tecton__proto_dot_common_dot_fco__locator__pb2._FCOLOCATOR
_FVMATERIALIZATION.fields_by_name['materialized_data_location'].message_type = _S3LOCATION
_FVMATERIALIZATION.fields_by_name['streaming_checkpoint_location'].message_type = _S3LOCATION
DESCRIPTOR.message_types_by_name['S3Location'] = _S3LOCATION
DESCRIPTOR.message_types_by_name['FvMaterialization'] = _FVMATERIALIZATION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

S3Location = _reflection.GeneratedProtocolMessageType('S3Location', (_message.Message,), {
  'DESCRIPTOR' : _S3LOCATION,
  '__module__' : 'tecton_proto.data.fv_materialization_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.S3Location)
  })
_sym_db.RegisterMessage(S3Location)

FvMaterialization = _reflection.GeneratedProtocolMessageType('FvMaterialization', (_message.Message,), {
  'DESCRIPTOR' : _FVMATERIALIZATION,
  '__module__' : 'tecton_proto.data.fv_materialization_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.FvMaterialization)
  })
_sym_db.RegisterMessage(FvMaterialization)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/materialization_status.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/data/materialization_status.proto',
  package='tecton_proto.data',
  syntax='proto2',
  serialized_options=b'\n\017com.tecton.dataP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n.tecton_proto/data/materialization_status.proto\x12\x11tecton_proto.data\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1ctecton_proto/common/id.proto\"\x8e\t\n\x1cMaterializationAttemptStatus\x12K\n\x10\x64\x61ta_source_type\x18\x01 \x01(\x0e\x32!.tecton_proto.data.DataSourceTypeR\x0e\x64\x61taSourceType\x12\x46\n\x11window_start_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0fwindowStartTime\x12\x42\n\x0fwindow_end_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\rwindowEndTime\x12H\n\x12\x66\x65\x61ture_start_time\x18\x0e \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x10\x66\x65\x61tureStartTime\x12\x44\n\x10\x66\x65\x61ture_end_time\x18\x0f \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0e\x66\x65\x61tureEndTime\x12\x62\n\x15materialization_state\x18\x04 \x01(\x0e\x32-.tecton_proto.data.MaterializationStatusStateR\x14materializationState\x12#\n\rstate_message\x18\n \x01(\tR\x0cstateMessage\x12-\n\x12termination_reason\x18\x0b \x01(\tR\x11terminationReason\x12\x32\n\x15spot_instance_failure\x18\x0c \x01(\x08R\x13spotInstanceFailure\x12O\n\x17materialization_task_id\x18\x05 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x15materializationTaskId\x12\x61\n\x1fmaterialization_task_created_at\x18\r \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x1cmaterializationTaskCreatedAt\x12%\n\x0e\x61ttempt_number\x18\x06 \x01(\x05R\rattemptNumber\x12I\n!spark_cluster_environment_version\x18\x07 \x01(\x05R\x1esparkClusterEnvironmentVersion\x12 \n\x0crun_page_url\x18\x08 \x01(\tR\nrunPageUrl\x12H\n\x12\x61ttempt_created_at\x18\t \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x10\x61ttemptCreatedAt\x12\x30\n\x14is_permanent_failure\x18\x10 \x01(\x08R\x12isPermanentFailure\x12\x39\n\nretry_time\x18\x11 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tretryTime\x12\x1a\n\x08progress\x18\x12 \x01(\x02R\x08progress\"\xca\x01\n\x15MaterializationStatus\x12\x45\n\x12\x66\x65\x61ture_package_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x10\x66\x65\x61turePackageId\x12j\n\x18materialization_attempts\x18\x02 \x03(\x0b\x32/.tecton_proto.data.MaterializationAttemptStatusR\x17materializationAttempts*\xa3\x01\n\x0e\x44\x61taSourceType\x12\x1c\n\x18\x44\x41TA_SOURCE_TYPE_UNKNOWN\x10\x00\x12\x1a\n\x16\x44\x41TA_SOURCE_TYPE_BATCH\x10\x01\x12\x1b\n\x17\x44\x41TA_SOURCE_TYPE_STREAM\x10\x02\x12\x1b\n\x17\x44\x41TA_SOURCE_TYPE_INGEST\x10\x03\x12\x1d\n\x19\x44\x41TA_SOURCE_TYPE_DELETION\x10\x04*\xc4\x02\n\x1aMaterializationStatusState\x12(\n$MATERIALIZATION_STATUS_STATE_UNKNOWN\x10\x00\x12*\n&MATERIALIZATION_STATUS_STATE_SCHEDULED\x10\x01\x12(\n$MATERIALIZATION_STATUS_STATE_PENDING\x10\x02\x12(\n$MATERIALIZATION_STATUS_STATE_RUNNING\x10\x03\x12(\n$MATERIALIZATION_STATUS_STATE_SUCCESS\x10\x04\x12&\n\"MATERIALIZATION_STATUS_STATE_ERROR\x10\x05\x12*\n&MATERIALIZATION_STATUS_STATE_CANCELLED\x10\x06\x42\x13\n\x0f\x63om.tecton.dataP\x01'
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,tecton__proto_dot_common_dot_id__pb2.DESCRIPTOR,])

_DATASOURCETYPE = _descriptor.EnumDescriptor(
  name='DataSourceType',
  full_name='tecton_proto.data.DataSourceType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DATA_SOURCE_TYPE_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DATA_SOURCE_TYPE_BATCH', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DATA_SOURCE_TYPE_STREAM', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DATA_SOURCE_TYPE_INGEST', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DATA_SOURCE_TYPE_DELETION', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1507,
  serialized_end=1670,
)
_sym_db.RegisterEnumDescriptor(_DATASOURCETYPE)

DataSourceType = enum_type_wrapper.EnumTypeWrapper(_DATASOURCETYPE)
_MATERIALIZATIONSTATUSSTATE = _descriptor.EnumDescriptor(
  name='MaterializationStatusState',
  full_name='tecton_proto.data.MaterializationStatusState',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='MATERIALIZATION_STATUS_STATE_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MATERIALIZATION_STATUS_STATE_SCHEDULED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MATERIALIZATION_STATUS_STATE_PENDING', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MATERIALIZATION_STATUS_STATE_RUNNING', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MATERIALIZATION_STATUS_STATE_SUCCESS', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MATERIALIZATION_STATUS_STATE_ERROR', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MATERIALIZATION_STATUS_STATE_CANCELLED', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1673,
  serialized_end=1997,
)
_sym_db.RegisterEnumDescriptor(_MATERIALIZATIONSTATUSSTATE)

MaterializationStatusState = enum_type_wrapper.EnumTypeWrapper(_MATERIALIZATIONSTATUSSTATE)
DATA_SOURCE_TYPE_UNKNOWN = 0
DATA_SOURCE_TYPE_BATCH = 1
DATA_SOURCE_TYPE_STREAM = 2
DATA_SOURCE_TYPE_INGEST = 3
DATA_SOURCE_TYPE_DELETION = 4
MATERIALIZATION_STATUS_STATE_UNKNOWN = 0
MATERIALIZATION_STATUS_STATE_SCHEDULED = 1
MATERIALIZATION_STATUS_STATE_PENDING = 2
MATERIALIZATION_STATUS_STATE_RUNNING = 3
MATERIALIZATION_STATUS_STATE_SUCCESS = 4
MATERIALIZATION_STATUS_STATE_ERROR = 5
MATERIALIZATION_STATUS_STATE_CANCELLED = 6



_MATERIALIZATIONATTEMPTSTATUS = _descriptor.Descriptor(
  name='MaterializationAttemptStatus',
  full_name='tecton_proto.data.MaterializationAttemptStatus',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='data_source_type', full_name='tecton_proto.data.MaterializationAttemptStatus.data_source_type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='dataSourceType', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='window_start_time', full_name='tecton_proto.data.MaterializationAttemptStatus.window_start_time', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='windowStartTime', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='window_end_time', full_name='tecton_proto.data.MaterializationAttemptStatus.window_end_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='windowEndTime', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_start_time', full_name='tecton_proto.data.MaterializationAttemptStatus.feature_start_time', index=3,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='featureStartTime', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_end_time', full_name='tecton_proto.data.MaterializationAttemptStatus.feature_end_time', index=4,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='featureEndTime', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='materialization_state', full_name='tecton_proto.data.MaterializationAttemptStatus.materialization_state', index=5,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='materializationState', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state_message', full_name='tecton_proto.data.MaterializationAttemptStatus.state_message', index=6,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='stateMessage', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='termination_reason', full_name='tecton_proto.data.MaterializationAttemptStatus.termination_reason', index=7,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='terminationReason', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='spot_instance_failure', full_name='tecton_proto.data.MaterializationAttemptStatus.spot_instance_failure', index=8,
      number=12, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='spotInstanceFailure', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='materialization_task_id', full_name='tecton_proto.data.MaterializationAttemptStatus.materialization_task_id', index=9,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='materializationTaskId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='materialization_task_created_at', full_name='tecton_proto.data.MaterializationAttemptStatus.materialization_task_created_at', index=10,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='materializationTaskCreatedAt', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='attempt_number', full_name='tecton_proto.data.MaterializationAttemptStatus.attempt_number', index=11,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='attemptNumber', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='spark_cluster_environment_version', full_name='tecton_proto.data.MaterializationAttemptStatus.spark_cluster_environment_version', index=12,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='sparkClusterEnvironmentVersion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run_page_url', full_name='tecton_proto.data.MaterializationAttemptStatus.run_page_url', index=13,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='runPageUrl', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='attempt_created_at', full_name='tecton_proto.data.MaterializationAttemptStatus.attempt_created_at', index=14,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='attemptCreatedAt', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_permanent_failure', full_name='tecton_proto.data.MaterializationAttemptStatus.is_permanent_failure', index=15,
      number=16, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='isPermanentFailure', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='retry_time', full_name='tecton_proto.data.MaterializationAttemptStatus.retry_time', index=16,
      number=17, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='retryTime', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='progress', full_name='tecton_proto.data.MaterializationAttemptStatus.progress', index=17,
      number=18, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='progress', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=133,
  serialized_end=1299,
)


_MATERIALIZATIONSTATUS = _descriptor.Descriptor(
  name='MaterializationStatus',
  full_name='tecton_proto.data.MaterializationStatus',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='feature_package_id', full_name='tecton_proto.data.MaterializationStatus.feature_package_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='featurePackageId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='materialization_attempts', full_name='tecton_proto.data.MaterializationStatus.materialization_attempts', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='materializationAttempts', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1302,
  serialized_end=1504,
)

_MATERIALIZATIONATTEMPTSTATUS.fields_by_name['data_source_type'].enum_type = _DATASOURCETYPE
_MATERIALIZATIONATTEMPTSTATUS.fields_by_name['window_start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_MATERIALIZATIONATTEMPTSTATUS.fields_by_name['window_end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_MATERIALIZATIONATTEMPTSTATUS.fields_by_name['feature_start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_MATERIALIZATIONATTEMPTSTATUS.fields_by_name['feature_end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_MATERIALIZATIONATTEMPTSTATUS.fields_by_name['materialization_state'].enum_type = _MATERIALIZATIONSTATUSSTATE
_MATERIALIZATIONATTEMPTSTATUS.fields_by_name['materialization_task_id'].message_type = tecton__proto_dot_common_dot_id__pb2._ID
_MATERIALIZATIONATTEMPTSTATUS.fields_by_name['materialization_task_created_at'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_MATERIALIZATIONATTEMPTSTATUS.fields_by_name['attempt_created_at'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_MATERIALIZATIONATTEMPTSTATUS.fields_by_name['retry_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_MATERIALIZATIONSTATUS.fields_by_name['feature_package_id'].message_type = tecton__proto_dot_common_dot_id__pb2._ID
_MATERIALIZATIONSTATUS.fields_by_name['materialization_attempts'].message_type = _MATERIALIZATIONATTEMPTSTATUS
DESCRIPTOR.message_types_by_name['MaterializationAttemptStatus'] = _MATERIALIZATIONATTEMPTSTATUS
DESCRIPTOR.message_types_by_name['MaterializationStatus'] = _MATERIALIZATIONSTATUS
DESCRIPTOR.enum_types_by_name['DataSourceType'] = _DATASOURCETYPE
DESCRIPTOR.enum_types_by_name['MaterializationStatusState'] = _MATERIALIZATIONSTATUSSTATE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

MaterializationAttemptStatus = _reflection.GeneratedProtocolMessageType('MaterializationAttemptStatus', (_message.Message,), {
  'DESCRIPTOR' : _MATERIALIZATIONATTEMPTSTATUS,
  '__module__' : 'tecton_proto.data.materialization_status_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.MaterializationAttemptStatus)
  })
_sym_db.RegisterMessage(MaterializationAttemptStatus)

MaterializationStatus = _reflection.GeneratedProtocolMessageType('MaterializationStatus', (_message.Message,), {
  'DESCRIPTOR' : _MATERIALIZATIONSTATUS,
  '__module__' : 'tecton_proto.data.materialization_status_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.MaterializationStatus)
  })
_sym_db.RegisterMessage(MaterializationStatus)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)
